from typing import Optional

from ninja import NinjaAPI

from django.test import TestCase

from app_utils.testing import create_user_from_evecharacter

from ledger.api.corporation.journal import LedgerJournalApiEndpoints
from ledger.models.corporationaudit import CorporationWalletJournalEntry
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all


class ManageApiJournalCorpEndpointsTest(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()

        cls.user, _ = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
                "ledger.corp_audit_manager",
                "ledger.corp_audit_admin_manager",
            ],
        )
        cls.user2, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "ledger.basic_access",
            ],
        )

        cls.api = NinjaAPI()
        cls.manage_api_endpoints = LedgerJournalApiEndpoints(api=cls.api)

    def test_get_corporation_journal_api_single(self):
        self.client.force_login(self.user)
        url = "/ledger/api/corporation/2001/wallet/"

        wallet_journal = (
            CorporationWalletJournalEntry.get_visible(self.user)
            .filter(division__corporation__corporation__corporation_id=2001)
            .select_related("first_party", "second_party", "division")
            .order_by("-entry_id")
        )

        start_count = (1 - 1) * 10000
        end_count = 1 * 10000

        wallet_journal = wallet_journal[start_count:end_count]

        response = self.client.get(url)

        expected_data = []
        for journal in wallet_journal:
            expected_data.append(
                {
                    "division": f"{journal.division.division} {journal.division.name}",
                    "id": journal.entry_id,
                    "date": journal.date.strftime("%Y-%m-%dT%H:%M:%SZ"),
                    "first_party": {
                        "id": journal.first_party.eve_id,
                        "name": journal.first_party.name,
                        "cat": journal.first_party.category,
                    },
                    "second_party": {
                        "id": journal.second_party.eve_id,
                        "name": journal.second_party.name,
                        "cat": journal.second_party.category,
                    },
                    "ref_type": journal.ref_type,
                    "balance": float(journal.balance),
                    "amount": float(journal.amount),
                    "reason": journal.reason,
                }
            )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_corporation_journal_api_single_ref_types(self):
        self.client.force_login(self.user)
        url = "/ledger/api/corporation/2001/wallet/?type_refs=bounty_prizes,ess_escrow_transfer"

        wallet_journal = (
            CorporationWalletJournalEntry.get_visible(self.user)
            .filter(division__corporation__corporation__corporation_id=2001)
            .select_related("first_party", "second_party", "division")
            .order_by("-entry_id")
        )

        start_count = (1 - 1) * 10000
        end_count = 1 * 10000

        wallet_journal = wallet_journal[start_count:end_count]

        response = self.client.get(url)

        expected_data = []
        for journal in wallet_journal:
            expected_data.append(
                {
                    "division": f"{journal.division.division} {journal.division.name}",
                    "id": journal.entry_id,
                    "date": journal.date.strftime("%Y-%m-%dT%H:%M:%SZ"),
                    "first_party": {
                        "id": journal.first_party.eve_id,
                        "name": journal.first_party.name,
                        "cat": journal.first_party.category,
                    },
                    "second_party": {
                        "id": journal.second_party.eve_id,
                        "name": journal.second_party.name,
                        "cat": journal.second_party.category,
                    },
                    "ref_type": journal.ref_type,
                    "balance": float(journal.balance),
                    "amount": float(journal.amount),
                    "reason": journal.reason,
                }
            )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_corporation_journal_api_single_ref_types_empty(self):
        self.client.force_login(self.user)
        url = (
            "/ledger/api/corporation/2001/wallet/?type_refs=12341234,,1234,1234,,,2134"
        )

        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

    def test_get_corporation_journal_api_no_permission(self):
        self.client.force_login(self.user2)
        url = "/ledger/api/corporation/2001/wallet/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 403)



""" AA Hooks"""

import logging

from ledger import app_settings
from ledger.errors import LedgerImportError


def get_extension_logger(name):
    """
    Takes the name of a plugin/extension and generates a child logger of the extensions logger
    to be used by the extension to log events to the extensions logger.

    The logging level is determined by the level defined for the parent logger.

    :param: name: the name of the extension doing the logging
    :return: an extensions child logger
    """

    logger_name = "ledger" if app_settings.LEDGER_LOGGER_USE else "extensions"

    if not isinstance(name, str):
        raise TypeError(
            f"get_extension_logger takes an argument of type string."
            f"Instead received argument of type {type(name).__name__}."
        )

    parent_logger = logging.getLogger(logger_name)

    logger = logging.getLogger(logger_name + "." + name)
    logger.name = name
    logger.level = parent_logger.level

    return logger


# pylint: disable=import-outside-toplevel
def get_corp_models_and_string():
    if app_settings.LEDGER_CORPSTATS_TWO:
        try:
            from corpstats.models import CorpMember

            return CorpMember
        except ImportError as exc:
            raise LedgerImportError("Corpstats is enabled but not installed") from exc

    from allianceauth.corputils.models import CorpMember

    return CorpMember



from typing import Optional

from ninja import NinjaAPI

from django.db.models import Q
from django.test import TestCase

from app_utils.testing import create_user_from_evecharacter

from ledger import app_settings
from ledger.api.character.journal import LedgerJournalApiEndpoints
from ledger.api.schema import Character
from ledger.models.characteraudit import CharacterWalletJournalEntry
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all


class ManageApiJournalCharEndpointsTest(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()

        cls.user, _ = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
                "ledger.char_audit_manager",
                "ledger.char_audit_admin_manager",
            ],
        )

        cls.user2, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "ledger.basic_access",
            ],
        )
        cls.api = NinjaAPI()
        cls.manage_api_endpoints = LedgerJournalApiEndpoints(api=cls.api)

    def test_get_character_journal_api(self):
        self.client.force_login(self.user)
        url = "/ledger/api/account/0/wallet/"

        response = self.client.get(url)

        journal_query = CharacterWalletJournalEntry.objects.filter(
            character__character__character_id=1001
        ).select_related("first_party", "second_party")
        expected_data = []
        for journal in journal_query:
            expected_data.append(
                {
                    "character": dict(
                        Character(**journal.character.character.__dict__)
                    ),
                    "id": journal.entry_id,
                    "date": journal.date.strftime("%Y-%m-%dT%H:%M:%SZ"),
                    "first_party": {
                        "id": (journal.first_party.eve_id),
                        "name": journal.first_party.name,
                        "cat": journal.first_party.category,
                    },
                    "second_party": {
                        "id": (journal.second_party.eve_id),
                        "name": journal.second_party.name,
                        "cat": journal.second_party.category,
                    },
                    "ref_type": journal.ref_type,
                    "balance": float(journal.balance),
                    "amount": float(journal.amount),
                    "reason": journal.reason,
                }
            )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_journal_api_single(self):
        self.client.force_login(self.user)
        url = "/ledger/api/account/1001/wallet/"

        journal_query = CharacterWalletJournalEntry.objects.filter(
            character__character__character_id=1001
        ).select_related("first_party", "second_party")

        response = self.client.get(url)

        expected_data = []
        for journal in journal_query:
            expected_data.append(
                {
                    "character": dict(
                        Character(**journal.character.character.__dict__)
                    ),
                    "id": journal.entry_id,
                    "date": journal.date.strftime("%Y-%m-%dT%H:%M:%SZ"),
                    "first_party": {
                        "id": (journal.first_party.eve_id),
                        "name": journal.first_party.name,
                        "cat": journal.first_party.category,
                    },
                    "second_party": {
                        "id": (journal.second_party.eve_id),
                        "name": journal.second_party.name,
                        "cat": journal.second_party.category,
                    },
                    "ref_type": journal.ref_type,
                    "balance": float(journal.balance),
                    "amount": float(journal.amount),
                    "reason": journal.reason,
                }
            )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_journal_api_single_ref_types(self):
        self.client.force_login(self.user)
        url = "/ledger/api/account/1001/wallet/?type_refs=bounty_prizes,player_donation,contract_reward,transaction_tax,industry_job_tax,market_escrow,insurance"

        journal_query = CharacterWalletJournalEntry.objects.filter(
            character__character__character_id=1001
        ).select_related("first_party", "second_party")

        response = self.client.get(url)

        expected_data = []
        for journal in journal_query:
            expected_data.append(
                {
                    "character": dict(
                        Character(**journal.character.character.__dict__)
                    ),
                    "id": journal.entry_id,
                    "date": journal.date.strftime("%Y-%m-%dT%H:%M:%SZ"),
                    "first_party": {
                        "id": (journal.first_party.eve_id),
                        "name": journal.first_party.name,
                        "cat": journal.first_party.category,
                    },
                    "second_party": {
                        "id": (journal.second_party.eve_id),
                        "name": journal.second_party.name,
                        "cat": journal.second_party.category,
                    },
                    "ref_type": journal.ref_type,
                    "balance": float(journal.balance),
                    "amount": float(journal.amount),
                    "reason": journal.reason,
                }
            )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_journal_api_single_ref_types_empty(self):
        self.client.force_login(self.user)
        url = "/ledger/api/account/1001/wallet/?type_refs=12341234,,1234,1234,,,2134"

        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

    def test_get_character_journal_api_no_permission(self):
        self.client.force_login(self.user2)
        url = "/ledger/api/account/1001/wallet/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 403)



from django.db.models import Q

from ledger.api.helpers import convert_ess_payout, get_alts_queryset
from ledger.api.managers.billboard_manager import BillboardData, BillboardLedger
from ledger.api.managers.core_manager import (
    LedgerDate,
    LedgerFilter,
    LedgerModels,
    LedgerTotal,
)
from ledger.hooks import get_extension_logger
from ledger.models.characteraudit import (
    CharacterMiningLedger,
    CharacterWalletJournalEntry,
)
from ledger.models.corporationaudit import CorporationWalletJournalEntry

logger = get_extension_logger(__name__)


class CharacterProcess:
    """JournalProcess class to process the journal entries."""

    def __init__(self, chars, year, month, corporations=None):
        self.year = year
        self.month = month
        self.chars = chars
        self.alts = self.get_alts(chars)
        self.chars_list = []
        self.corp = corporations if corporations else []
        self.summary_total = LedgerTotal()

    def get_alts(self, main):
        """Get the Alts for the Main Character"""
        alts = None
        if len(main) == 1:
            # pylint: disable=broad-exception-caught
            try:
                alts = get_alts_queryset(main[0])
            except Exception:
                pass
        return alts

    def process_character_chars(self, journal):
        """Process the characters for the Journal"""
        character_dict = {}
        character_totals = LedgerTotal()

        # Step 1: Annotated the Journal Entries
        def process_char(char):
            char_name = char.character_name
            char_id = char.character_id

            # Call annotate_ledger and store the result
            result = journal.generate_ledger(
                [char_id], self.filter_date, self.chars_list
            )

            amounts = result["amounts"]
            amounts_others = result["amounts_others"]
            amounts_costs = result["amounts_costs"]

            # Convert the ESS Payout for Character
            amounts["ess"] = convert_ess_payout(amounts["ess"])

            # Summing amounts_others
            total_amount_others = sum(amounts_others.values())

            # Summing amounts_costs
            costs_amount = sum(amounts_costs.values())

            summary_amount = (
                amounts["bounty"]
                + amounts["ess"]
                + amounts["mining"]
                + total_amount_others
            )

            summary_amount -= abs(costs_amount)

            if summary_amount:
                character_dict[char_id] = {
                    "main_id": char_id,
                    "main_name": char_name,
                    "total_amount": amounts["bounty"],
                    "total_amount_ess": amounts["ess"],
                    "total_amount_mining": amounts["mining"],
                    "total_amount_others": total_amount_others,
                    "total_amount_costs": costs_amount,
                }

            totals = {
                "total_amount": amounts["bounty"],
                "total_amount_ess": amounts["ess"],
                "total_amount_all": summary_amount,
                "total_amount_mining": amounts["mining"],
                "total_amount_others": total_amount_others,
                "total_amount_costs": costs_amount,
            }
            # Summary all
            character_totals.get_data(totals)

        for char in self.chars:
            process_char(char)

        return character_dict, character_totals.to_dict()

    def generate_ledger(self):
        # Get the Character IDs
        if self.alts:
            self.chars_list = [char.character_id for char in self.alts]
        else:
            self.chars_list = [char.character_id for char in self.chars]

        self.filter_date = Q(date__year=self.year)
        if not self.month == 0:
            self.filter_date &= Q(date__month=self.month)

        # Filter for the Character Journal
        journal = CharacterWalletJournalEntry.objects.filter(
            self.filter_date
        ).select_related("first_party", "second_party")

        # Create the Dicts for each Character
        character_dict, character_totals = self.process_character_chars(journal)

        output = []
        output.append(
            {
                "ratting": sorted(
                    list(character_dict.values()), key=lambda x: x["main_name"]
                ),
                "total": character_totals,
            }
        )

        return output

    def generate_billboard(self):
        # Get the Character IDs
        if self.alts:
            self.chars_list = [char.character_id for char in self.alts]
        else:
            self.chars_list = [char.character_id for char in self.chars]

        # Get the Filter Settings
        filters = LedgerFilter(self.chars_list)

        filter_date = Q(date__year=self.year)
        if not self.month == 0:
            filter_date &= Q(date__month=self.month)

        # Filter for the Character Journal
        journal = CharacterWalletJournalEntry.objects.filter(
            filters.filter_partys, filter_date
        )

        corporation_journal = CorporationWalletJournalEntry.objects.filter(
            filters.filter_second_party, filter_date
        )

        mining_journal = CharacterMiningLedger.objects.filter(
            filters.filter_mining, filter_date
        ).annotate_pricing()

        # Create Data for Billboard
        date_data = LedgerDate(self.year, self.month)
        data = BillboardData()
        models = LedgerModels(
            character_journal=journal,
            corporation_journal=corporation_journal,
            mining_journal=mining_journal,
        )

        # Create the Billboard for the Characters
        ledger = BillboardLedger(date_data, models, data, corp=False)
        billboard_dict = ledger.billboard_char_ledger(self.chars, self.chars_list)

        output = []
        output.append(
            {
                "billboard": billboard_dict,
            }
        )

        return output



"""PvE Views"""

# Django
from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import render

from ledger.hooks import get_extension_logger

# Ledger
from ledger.view_helpers.core import add_info_to_context

logger = get_extension_logger(__name__)


@login_required
@permission_required("ledger.basic_access")
def ledger_index(request):
    context = {}
    context = add_info_to_context(request, context)

    return render(request, "ledger/index.html", context=context)



from .characteraudit import (
    CharacterAudit,
    CharacterMiningLedger,
    CharacterWalletJournalEntry,
)
from .corporationaudit import (
    CorporationAudit,
    CorporationWalletDivision,
    CorporationWalletJournalEntry,
)
from .events import Events
from .general import EveEntity, General



from unittest.mock import MagicMock, patch

from django.core.cache import cache
from django.db.models import QuerySet
from django.test import TestCase

from ledger.models.events import Events
from ledger.view_helpers.core import (
    _storage_key,
    add_info_to_context,
    calculate_ess_stolen,
    delete_cache,
    events_filter,
    get_cache_stale,
    ledger_cache_timeout,
    set_cache,
    set_cache_hourly,
)

MODULE_PATH = "ledger.view_helpers.core"


class TestViewHelpers(TestCase):
    def test_calculate_ess_stolen_key_error(self):
        # Test case where 'bounty' key is missing
        amounts = {"ess": {"total_amount": 1000, "total_amount_day": 500}, "stolen": {}}

        # Call the function and check if it handles the KeyError
        result = calculate_ess_stolen(amounts)

        # Since 'bounty' key is missing, 'stolen' amounts should not be calculated
        self.assertEqual(result["stolen"].get("total_amount"), None)
        self.assertEqual(result["stolen"].get("total_amount_day"), None)

    def test_calculate_ess_stolen_key_error_ess(self):
        # Test case where 'ess' key is missing
        amounts = {
            "bounty": {"total_amount": 1000, "total_amount_day": 500},
            "stolen": {},
        }

        # Call the function and check if it handles the KeyError
        result = calculate_ess_stolen(amounts)

        # Since 'ess' key is missing, 'stolen' amounts should not be calculated
        self.assertEqual(result["stolen"].get("total_amount"), None)
        self.assertEqual(result["stolen"].get("total_amount_day"), None)

    def test_events_filter(self):
        result = events_filter([])
        self.assertEqual(result, [])

    def test_get_cache_stale(self):
        result = get_cache_stale("test")
        self.assertFalse(result)

    @patch.object(cache, "get", return_value="test_value")
    def test_get_cache_stale_with_data(self, mock_cache_get):
        # given
        key_name = "test_key"
        # when
        result = get_cache_stale(key_name)
        # then
        mock_cache_get.assert_called_once_with(key=_storage_key(key_name))
        self.assertEqual(result, "test_value")

    def test_set_cache(self):
        result = set_cache("test", "test", 1)
        self.assertIsNone(result)

    def test_set_cache_hourly(self):
        result = set_cache_hourly("test", "test")
        self.assertIsNone(result)

    def test_delete_cache(self):
        result = delete_cache("test")
        self.assertIsNone(result)

    def test_storage_key(self):
        result = _storage_key("test")
        self.assertEqual(result, "ledger_storage_test")


class TestEventsFilter(TestCase):
    def setUp(self):
        self.entries = MagicMock(spec=QuerySet)
        self.excluded_entries = MagicMock(spec=QuerySet)
        self.entries.exclude.return_value = self.excluded_entries

    @patch.object(Events, "objects")
    def test_events_filter_no_char_ledger(self, mock_objects):
        # Erstellen Sie ein Event ohne char_ledger
        event = MagicMock(char_ledger=None)
        mock_objects.all.return_value = [event]

        # when
        result = events_filter(self.entries)

        # then
        self.entries.exclude.assert_not_called()
        self.assertEqual(result, self.entries)

    @patch.object(Events, "objects")
    def test_events_filter_with_q_objects(self, mock_objects):
        # Erstellen Sie ein Event mit char_ledger und einem Datumsbereich
        event = MagicMock(
            char_ledger=True, date_start="2022-01-01", date_end="2022-01-31"
        )
        mock_objects.all.return_value = [event]

        # when
        result = events_filter(self.entries)

        # then
        self.entries.exclude.assert_called_once()
        self.assertEqual(result, self.excluded_entries)

    @patch.object(Events, "objects")
    def test_events_filter_multiple(self, mock_objects):
        # Erstellen Sie mehrere Events, von denen mindestens eines char_ledger und einen gültigen Datumsbereich hat
        event1 = MagicMock(
            char_ledger=True, date_start="2022-01-01", date_end="2022-01-31"
        )
        event2 = MagicMock(
            char_ledger=False, date_start="2022-02-01", date_end="2022-02-28"
        )
        event3 = MagicMock(
            char_ledger=True, date_start="2022-03-01", date_end="2022-03-31"
        )
        mock_objects.all.return_value = [event1, event2, event3]

        # when
        result = events_filter(self.entries)

        # then
        self.entries.exclude.assert_called_once()
        self.assertEqual(result, self.entries.exclude.return_value)


class TestCache(TestCase):
    def test_ledger_cache_timeout(self):
        # when
        result = ledger_cache_timeout()
        # then
        self.assertTrue(result > 0)

    def test_add_info_to_context(self):
        # given
        request = MagicMock()
        request.user = MagicMock()
        request.user.id = 99999999

        context = {"theme": None}
        # when
        result = add_info_to_context(request, context)
        # then
        self.assertEqual(result, context)



from datetime import timedelta
from unittest.mock import patch

from django.test import TestCase
from django.utils import timezone

from ledger.models.events import Events

MODULE_PATH = "ledger.models.general"


class TestGeneralModel(TestCase):
    def setUp(self):
        self.events = Events()
        self.events.id = 1
        self.events.title = "Test Title"
        self.events.date_start = timezone.now()
        self.events.date_end = timezone.now() + timedelta(days=1)
        self.events.description = "Test Description"
        self.events.char_ledger = True
        self.events.location = "Test Location"

    def test_str(self):
        self.assertEqual(str(self.events), "Event 1")



"""Generate AllianceAuth test objects from allianceauth.json."""

import json
from pathlib import Path

from memberaudit.models import Character

from allianceauth.eveonline.models import EveCharacter


def create_memberaudit(eve_character: EveCharacter, **kwargs) -> Character:
    """Create a memberaudit object for the given character."""
    params = {"eve_character": eve_character}
    params.update(kwargs)
    obj = Character(**params)
    obj.save(ignore_cache=True)
    return obj


def load_memberaudit():
    Character.objects.all().delete()
    create_memberaudit(EveCharacter.objects.get(character_id=1001), is_disabled=False)
    create_memberaudit(EveCharacter.objects.get(character_id=1002), is_disabled=False)
    create_memberaudit(EveCharacter.objects.get(character_id=1003), is_disabled=False)



from django.db import models
from django.db.models import ExpressionWrapper, F

from allianceauth.eveonline.models import EveCharacter

from ledger.hooks import get_extension_logger

logger = get_extension_logger(__name__)


class AuditCharacterQuerySet(models.QuerySet):
    # pylint: disable=duplicate-code
    def visible_to(self, user):
        # superusers get all visible
        if user.is_superuser:
            logger.debug("Returning all characters for superuser %s.", user)
            return self

        if user.has_perm("ledger.char_audit_admin_manager"):
            logger.debug("Returning all characters for %s.", user)
            return self

        try:
            char = user.profile.main_character
            assert char
            queries = [models.Q(character__character_ownership__user=user)]

            if user.has_perm("ledger.char_audit_manager"):
                queries.append(models.Q(character__corporation_id=char.corporation_id))

            logger.debug(
                "%s queries for user %s visible chracters.", len(queries), user
            )

            query = queries.pop()
            for q in queries:
                query |= q
            return self.filter(query)
        except AssertionError:
            logger.debug("User %s has no main character. Nothing visible.", user)
            return self.none()


class AuditCharacterManager(models.Manager):
    def get_queryset(self):
        return AuditCharacterQuerySet(self.model, using=self._db)

    @staticmethod
    def visible_eve_characters(user):
        qs = EveCharacter.objects.get_queryset()
        if user.is_superuser:
            logger.debug("Returning all characters for superuser %s.", user)
            return qs.all()

        if user.has_perm("ledger.char_audit_admin_manager"):
            logger.debug("Returning all characters for %s.", user)
            return qs.all()

        try:
            char = user.profile.main_character
            assert char
            queries = [models.Q(character_ownership__user=user)]

            if user.has_perm("ledger.char_audit_manager"):
                queries.append(models.Q(corporation_id=char.corporation_id))

            logger.debug(
                "%s queries for user %s visible chracters.", len(queries), user
            )

            query = queries.pop()
            for q in queries:
                query |= q
            return qs.filter(query)
        except AssertionError:
            logger.debug("User %s has no main character. Nothing visible.", user)
            return qs.none()

    def visible_to(self, user):
        return self.get_queryset().visible_to(user)


class CharacterMiningLedgerEntryQueryset(models.QuerySet):
    def annotate_pricing(self) -> models.QuerySet:
        """Annotate price and total columns."""
        return (
            self.select_related("type__market_price")
            .annotate(price=F("type__market_price__average_price"))
            .annotate(
                total=ExpressionWrapper(
                    F("type__market_price__average_price") * F("quantity"),
                    output_field=models.FloatField(),
                ),
            )
        )


class CharacterMiningLedgerEntryManagerBase(models.Manager):
    pass


CharacterMiningLedgerEntryManager = CharacterMiningLedgerEntryManagerBase.from_queryset(
    CharacterMiningLedgerEntryQueryset
)



from typing import List

from ninja import NinjaAPI

from django.shortcuts import render
from django.utils import timezone

from allianceauth.eveonline.models import EveCharacter

from ledger.api import schema
from ledger.api.helpers import get_alts_queryset, get_character
from ledger.api.managers.template_manager import TemplateData, TemplateProcess
from ledger.hooks import get_extension_logger

logger = get_extension_logger(__name__)


# pylint: disable=too-many-locals, too-many-branches, too-many-statements
class LedgerTemplateApiEndpoints:
    tags = ["CharacerLedgerTemplate"]

    def __init__(self, api: NinjaAPI):

        @api.get(
            "account/{character_id}/ledger/template/year/{year}/month/{month}/",
            response={200: List[schema.CharacterLedgerTemplate], 403: str},
            tags=self.tags,
        )
        def get_character_ledger_template(
            request, character_id: int, year: int, month: int
        ):
            overall_mode = character_id == 0
            # Set the character_id to the main character if the character_id is 0
            character_id = (
                request.user.profile.main_character.character_id
                if character_id == 0
                else character_id
            )
            current_date = timezone.now()

            perms, main = get_character(request, character_id)
            alts = get_alts_queryset(main)

            chars_list = [char.character_id for char in alts]

            if not perms:
                context = {
                    "error_title": "Permission Denied",
                    "error_message": "You don't have permission to view this character's ledger.",
                }
                return render(request, "ledger/modals/information/error.html", context)
            try:
                if overall_mode:
                    linked_char = EveCharacter.objects.filter(
                        character_id__in=chars_list,
                    )
                else:
                    linked_char = [
                        EveCharacter.objects.get(
                            character_id=character_id,
                        )
                    ]
            except EveCharacter.DoesNotExist:
                context = {
                    "error_title": "403 Error",
                    "error_message": "Character not found.",
                }
                return render(request, "ledger/modals/information/error.html", context)

            # Create the Ledger
            ledger_data = TemplateData(request, main, year, month, current_date)
            ledger = TemplateProcess(linked_char, ledger_data, overall_mode)
            context = {
                "character": ledger.character_template(),
                "mode": "Ratting",
            }

            return render(
                request,
                "ledger/modals/information/view_character_content.html",
                context,
            )



import datetime
from unittest.mock import PropertyMock, patch

from django.test import RequestFactory, TestCase
from django.utils import timezone

from allianceauth.eveonline.models import EveCharacter
from app_utils.testing import create_user_from_evecharacter

from ledger.app_settings import LEDGER_CHAR_MAX_INACTIVE_DAYS
from ledger.models.characteraudit import (
    CharacterAudit,
    CharacterMiningLedger,
    CharacterWalletJournalEntry,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all

MODULE_PATH = "ledger.models.general"


class TestCharacterAuditModel(TestCase):
    @classmethod
    def setUp(self):
        load_allianceauth()
        self.audit = CharacterAudit(
            character=EveCharacter.objects.get(character_id=1001)
        )

    def test_str(self):
        self.assertEqual(str(self.audit), "Gneuten's Character Data")

    def test_is_active(self):
        self.assertTrue(self.audit.is_active)

    def test_is_active_false(self):
        # given/when
        self.audit.last_update_wallet = timezone.now() - datetime.timedelta(
            days=LEDGER_CHAR_MAX_INACTIVE_DAYS + 1
        )
        self.audit.last_update_mining = timezone.now() - datetime.timedelta(
            days=LEDGER_CHAR_MAX_INACTIVE_DAYS + 1
        )
        self.audit.save()
        # then
        self.assertFalse(self.audit.is_active())

    def test_is_active_updates_active_field(self):
        # given/when
        self.audit.last_update_wallet = timezone.now() - datetime.timedelta(
            days=LEDGER_CHAR_MAX_INACTIVE_DAYS + 1
        )
        self.audit.last_update_mining = timezone.now() - datetime.timedelta(
            days=LEDGER_CHAR_MAX_INACTIVE_DAYS + 1
        )
        self.audit.active = True
        self.audit.save()
        # then
        self.assertFalse(self.audit.is_active())

        self.audit.refresh_from_db()
        self.assertFalse(self.audit.active)

    def test_is_active_exception(self):
        with patch.object(
            self.audit.__class__, "last_update_wallet", new_callable=PropertyMock
        ) as mock_wallet:
            with patch.object(
                self.audit.__class__, "last_update_mining", new_callable=PropertyMock
            ) as mock_mining:
                # Make the mocks raise an exception when accessed
                mock_wallet.side_effect = Exception
                mock_mining.side_effect = Exception

                self.assertFalse(self.audit.is_active())

    def test_get_esi_scopes(self):
        self.assertEqual(
            self.audit.get_esi_scopes(),
            [
                # Mining Ledger
                "esi-industry.read_character_mining.v1",
                # Wallet / Market /  Contracts
                "esi-wallet.read_character_wallet.v1",
                "esi-contracts.read_character_contracts.v1",
            ],
        )


class TestCharacterWalletJournal(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()
        cls.journal = CharacterWalletJournalEntry.objects.get(entry_id=1)

    def test_str(self):
        self.assertEqual(
            str(self.journal),
            "Character Wallet Journal: CONCORD 'bounty_prizes' Gneuten: 100000.00 isk",
        )

    def test_get_visible(self):
        self.factory = RequestFactory()
        self.user, self.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.char_audit_admin_manager",
            ],
        )
        request = self.factory.get("/")
        request.user = self.user

        query = CharacterWalletJournalEntry.get_visible(request.user)

        excepted_character = CharacterWalletJournalEntry.objects.all()

        self.assertEqual(list(query), list(excepted_character))


class TestCharacterMiningLedger(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()
        cls.mining = CharacterMiningLedger.objects.get(
            id="20240316-17425-1001-30004783"
        )

    def test_str(self):
        self.assertEqual(
            str(self.mining), "Gneuten's Character Data 20240316-17425-1001-30004783"
        )



from typing import List

from ninja import NinjaAPI

from django.shortcuts import render
from django.utils import timezone

from allianceauth.eveonline.models import EveCharacter

from ledger.api import schema
from ledger.api.helpers import (
    get_alts_queryset,
    get_character,
    get_main_and_alts_all,
    get_main_and_alts_corporations,
)
from ledger.api.managers.template_manager import TemplateData, TemplateProcess
from ledger.hooks import get_extension_logger

logger = get_extension_logger(__name__)


# pylint: disable=too-many-locals, too-many-branches, too-many-statements
class LedgerTemplateApiEndpoints:
    tags = ["CorporationLedgerTemplate"]

    def __init__(self, api: NinjaAPI):

        @api.get(
            "corporation/{corporation_id}/character/{main_id}/ledger/template/year/{year}/month/{month}/",
            response={200: List[schema.CharacterLedgerTemplate], 403: str},
            tags=self.tags,
        )
        def get_corporation_ledger_template(
            request,
            corporation_id: int,
            main_id: int,
            year: int,
            month: int,
            corp: bool = False,
        ):
            perms = request.user.has_perm("ledger.basic_access")

            if not perms:
                return 403, "Permission Denied"

            current_date = timezone.now()

            if corporation_id == 0:
                corporations = get_main_and_alts_corporations(request)
            else:
                corporations = [corporation_id]

            _, char = get_character(request, main_id)

            # Get all Chars from the main (including the main char itself)
            alts = get_alts_queryset(char, corporations=corporations)
            linked_char = list(alts)

            overall_mode = main_id == 0

            if overall_mode:
                _, chars_list = get_main_and_alts_all(corporations)
                linked_char = EveCharacter.objects.filter(
                    character_id__in=chars_list,
                )
            elif corp:
                linked_char = EveCharacter.objects.filter(
                    corporation_id__in=[main_id],
                )
                overall_mode = True

            # Create the Ledger

            ledger_data = TemplateData(request, char, year, month, current_date)
            ledger = TemplateProcess(linked_char, ledger_data, overall_mode)
            context = {
                "character": ledger.corporation_template(),
                "mode": "TAX",
            }

            return render(
                request,
                "ledger/modals/information/view_character_content.html",
                context,
            )



import sys
from unittest.mock import patch

from django.test import TestCase

from ledger.errors import LedgerImportError
from ledger.hooks import get_corp_models_and_string, get_extension_logger

MODULE_PATH = "ledger.hooks"


class TestTemplateTags(TestCase):
    def test_logger_fail(self):
        with self.assertRaises(TypeError):
            get_extension_logger(1234)


class TestApiHelperCorpStatsImport(TestCase):
    def setUp(self):
        self.original_sys_modules = sys.modules.copy()

    def tearDown(self):
        sys.modules = self.original_sys_modules

    @patch(MODULE_PATH + ".app_settings.LEDGER_CORPSTATS_TWO", True)
    def test_packages_are_not_installed(self):
        with patch.dict(
            sys.modules,
            {k: None for k in list(sys.modules) if k.startswith("corpstats")},
        ):
            with self.assertRaises(LedgerImportError):
                _ = get_corp_models_and_string()



"""
Character Audit Model
"""

import datetime
from typing import List

from django.db import models
from django.utils import timezone
from eveuniverse.models import EveSolarSystem, EveType

from allianceauth.eveonline.models import EveCharacter

from ledger import app_settings
from ledger.hooks import get_extension_logger
from ledger.managers.characterjournal_manager import CharWalletManager
from ledger.managers.charaudit_manager import (
    AuditCharacterManager,
    CharacterMiningLedgerEntryManager,
)
from ledger.models.general import EveEntity

logger = get_extension_logger(__name__)


class CharacterAudit(models.Model):

    id = models.AutoField(primary_key=True)

    active = models.BooleanField(default=False)

    character = models.OneToOneField(
        EveCharacter, on_delete=models.CASCADE, related_name="ledger_character"
    )

    last_update_wallet = models.DateTimeField(null=True, default=None, blank=True)

    last_update_mining = models.DateTimeField(null=True, default=None, blank=True)

    balance = models.DecimalField(
        max_digits=20, decimal_places=2, null=True, default=None
    )

    objects = AuditCharacterManager()

    def __str__(self):
        return f"{self.character.character_name}'s Character Data"

    class Meta:
        default_permissions = ()
        permissions = (
            ("char_audit_manager", "Has access to all characters for own Corp"),
            ("char_audit_admin_manager", "Has access to all characters"),
        )

    @classmethod
    def get_esi_scopes(cls) -> List[str]:
        """Return list of required ESI scopes to fetch."""
        return [
            # Mining Ledger
            "esi-industry.read_character_mining.v1",
            # Wallet / Market /  Contracts
            "esi-wallet.read_character_wallet.v1",
            "esi-contracts.read_character_contracts.v1",
        ]

    def is_active(self):
        time_ref = timezone.now() - datetime.timedelta(
            days=app_settings.LEDGER_CHAR_MAX_INACTIVE_DAYS
        )
        try:
            is_active = True

            is_active = self.last_update_wallet > time_ref
            is_active = self.last_update_mining > time_ref

            if self.active != is_active:
                self.active = is_active
                self.save()

            return is_active
        except Exception:  # pylint: disable=broad-exception-caught
            return False


class WalletJournalEntry(models.Model):
    amount = models.DecimalField(
        max_digits=20, decimal_places=2, null=True, default=None
    )
    balance = models.DecimalField(
        max_digits=20, decimal_places=2, null=True, default=None
    )
    context_id = models.BigIntegerField(null=True, default=None)

    class ContextType(models.TextChoices):
        STRUCTURE_ID = "structure_id"
        STATION_ID = "station_id"
        MARKET_TRANSACTION_ID = "market_transaction_id"
        CHARACTER_ID = "character_id"
        CORPORATION_ID = "corporation_id"
        ALLIANCE_ID = "alliance_id"
        EVE_SYSTEM = "eve_system"
        INDUSTRY_JOB_ID = "industry_job_id"
        CONTRACT_ID = "contract_id"
        PLANET_ID = "planet_id"
        SYSTEM_ID = "system_id"
        TYPE_ID = "type_id"

    context_id_type = models.CharField(
        max_length=30, choices=ContextType.choices, null=True, default=None
    )
    date = models.DateTimeField()
    description = models.CharField(max_length=500)
    first_party = models.ForeignKey(
        EveEntity,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        related_name="+",
    )
    entry_id = models.BigIntegerField()
    reason = models.CharField(max_length=500, null=True, default=None)
    ref_type = models.CharField(max_length=72)
    second_party = models.ForeignKey(
        EveEntity,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        related_name="+",
    )
    tax = models.DecimalField(max_digits=20, decimal_places=2, null=True, default=None)
    tax_receiver_id = models.IntegerField(null=True, default=None)

    class Meta:
        abstract = True
        indexes = (
            models.Index(fields=["date"]),
            models.Index(fields=["amount"]),
            models.Index(fields=["entry_id"]),
            models.Index(fields=["ref_type"]),
            models.Index(fields=["first_party"]),
            models.Index(fields=["second_party"]),
        )
        default_permissions = ()


class CharacterWalletJournalEntry(WalletJournalEntry):
    character = models.ForeignKey(CharacterAudit, on_delete=models.CASCADE)

    objects = CharWalletManager()

    def __str__(self):
        return f"Character Wallet Journal: {self.first_party.name} '{self.ref_type}' {self.second_party.name}: {self.amount} isk"

    @classmethod
    def get_visible(cls, user):
        chars_vis = CharacterAudit.objects.visible_to(user)
        return cls.objects.filter(character__in=chars_vis)


# Mining Models


class CharacterMiningLedger(models.Model):
    id = models.CharField(max_length=50, primary_key=True)
    character = models.ForeignKey(
        CharacterAudit, on_delete=models.CASCADE, related_name="ledger_character"
    )
    date = models.DateField()
    type = models.ForeignKey(
        EveType, on_delete=models.CASCADE, related_name="ledger_evetype"
    )
    system = models.ForeignKey(
        EveSolarSystem, on_delete=models.CASCADE, related_name="ledger_evesolarsystem"
    )
    quantity = models.IntegerField()

    @staticmethod
    def create_primary_key(character_id, mining_record):
        return f"{mining_record['date'].strftime('%Y%m%d')}-{mining_record['type_id']}-{character_id}-{mining_record['solar_system_id']}"

    objects = CharacterMiningLedgerEntryManager()

    class Meta:
        default_permissions = ()

    def __str__(self) -> str:
        return f"{self.character} {self.id}"



from unittest.mock import Mock, patch

from django.contrib.sessions.middleware import SessionMiddleware
from django.test import RequestFactory, TestCase
from django.urls import reverse
from django.utils import timezone
from esi.models import Token

from app_utils.testing import create_user_from_evecharacter

from ledger.models.corporationaudit import CorporationAudit
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_memberaudit import load_memberaudit
from ledger.views.corporation.corp_audit import add_corp

MODULE_PATH = "ledger.views.corporation.corp_audit"


class CharAuditTest(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
                "ledger.admin_access",
                "ledger.corp_audit_admin_manager",
            ],
        )

    @patch(MODULE_PATH + ".messages")
    @patch(MODULE_PATH + ".update_corp")
    def test_add_corp(self, mock_update_corp, mock_messages):
        self.client.force_login(self.user)
        token = Mock(spec=Token)
        token.character_id = self.character_ownership.character.character_id
        token.corporation_id = self.character_ownership.character.corporation_id
        token.corporation_name = self.character_ownership.character.corporation_name
        token.corporation_ticker = self.character_ownership.character.corporation_ticker
        request = self.factory.get(reverse("ledger:ledger_add_corp"))
        request.user = self.user
        request.token = token
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        # given
        orig_view = add_corp.__wrapped__.__wrapped__
        # when
        response = orig_view(request, token)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("ledger:ledger_index"))
        self.assertTrue(mock_messages.info.called)
        self.assertTrue(mock_update_corp.apply_async.called)
        self.assertTrue(
            CorporationAudit.objects.get(
                corporation=self.character_ownership.character.corporation
            )
        )
